/*
 * RegExpToStringComposer.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#include "RegExpToStringComposer.h"
#include "regexp/RegExpClasses.h"

#include "regexp/unbounded/UnboundedRegExpElements.h"
#include "regexp/formal/FormalRegExpElements.h"

#include "../StringApi.hpp"

namespace regexp {

void RegExpToStringComposer::compose(std::ostream& output, const UnboundedRegExp < > & regexp) {
	RegExpToStringComposer::compose ( output, regexp.getRegExp ( ) );
}

auto RegExpToStringComposerUnboundedRegExp = RegExpToStringComposer::RegistratorWrapper<void, UnboundedRegExp < > >(RegExpToStringComposer::compose);

void RegExpToStringComposer::compose(std::ostream& output, const UnboundedRegExpStructure < alphabet::Symbol > & regexp) {
	Priority tmp = Priority::ALTERNATION;
	std::tuple<Priority&, std::ostream&> out = std::tie(tmp, output);
	regexp.getStructure().accept<void, RegExpToStringComposer::Unbounded > (out);
}

void RegExpToStringComposer::compose(std::ostream& output, const FormalRegExp < > & regexp) {
	RegExpToStringComposer::compose ( output, regexp.getRegExp ( ) );
}

auto RegExpToStringComposerFormalRegExp = RegExpToStringComposer::RegistratorWrapper<void, FormalRegExp < > >(RegExpToStringComposer::compose);

void RegExpToStringComposer::compose(std::ostream& output, const FormalRegExpStructure < alphabet::Symbol > & regexp) {
	Priority tmp = Priority::ALTERNATION;
	std::tuple<Priority&, std::ostream&> out = std::tie(tmp, output);
	regexp.getStructure().accept<void, RegExpToStringComposer::Formal > (out);
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpAlternation < alphabet::Symbol > & alternation, std::tuple<Priority&, std::ostream &> & out ) {
	if(alternation.getElements().size() == 0) {
		std::get<1>(out) << "#0";
	} else if(alternation.getElements().size() == 1) {
		alternation.getElements()[0]->accept<void, RegExpToStringComposer::Unbounded > (out);
	} else {
		Priority outerPriorityMinimum = std::get<0>(out);
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';
		bool first = true;
		for (const auto& element : alternation.getElements()) {
			if(first) {
				first = false;
			} else {
				std::get<1>(out) << '+';
			}
			std::get<0>(out) = Priority::ALTERNATION;
			element->accept<void, RegExpToStringComposer::Unbounded > (out);
		}
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
	}
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpConcatenation < alphabet::Symbol > & concatenation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(concatenation.getElements().size() == 0) {
		std::get<1>(out) << "#E";
	} else if(concatenation.getElements().size() == 1) {
		concatenation.getElements()[0]->accept<void, RegExpToStringComposer::Unbounded > (out);
	} else {
		if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';
		bool first = true;
		for (auto element : concatenation.getElements()) {
			if(first) {
				first = false;
			} else {
				std::get<1>(out) << ' ';
			}
			std::get<0>(out) = Priority::CONCATENATION;
			element->accept<void, RegExpToStringComposer::Unbounded > (out);
		}
		if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
	}
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpIteration < alphabet::Symbol > & iteration, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<0>(out) = Priority::FACTOR;
	iteration.getElement().accept<void, RegExpToStringComposer::Unbounded > (out);
	std::get<1>(out) << "*";
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpSymbol < alphabet::Symbol > & symbol, std::tuple<Priority&, std::ostream &> & out ) {
	alib::stringApi<alphabet::Symbol>::compose(std::get<1>(out), symbol.getSymbol());
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpEpsilon < alphabet::Symbol > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#E";
}

void RegExpToStringComposer::Unbounded::visit( const UnboundedRegExpEmpty < alphabet::Symbol > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#0";
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpAlternation < alphabet::Symbol > & alternation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';

	std::get<0>(out) = Priority::ALTERNATION;
	alternation.getLeftElement().accept<void, RegExpToStringComposer::Formal > (out);
	std::get<1>(out) << '+';
	alternation.getRightElement().accept<void, RegExpToStringComposer::Formal > (out);

	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpConcatenation < alphabet::Symbol > & concatenation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';

	std::get<0>(out) = Priority::CONCATENATION;
	concatenation.getLeftElement().accept<void, RegExpToStringComposer::Formal > (out);
	std::get<1>(out) << ' ';
	concatenation.getRightElement().accept<void, RegExpToStringComposer::Formal > (out);

	if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpIteration < alphabet::Symbol > & iteration, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<0>(out) = Priority::FACTOR;
	iteration.getElement().accept<void, RegExpToStringComposer::Formal > (out);
	std::get<1>(out) << "*";
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpSymbol < alphabet::Symbol > & symbol, std::tuple<Priority&, std::ostream &> & out ) {
	alib::stringApi<alphabet::Symbol>::compose(std::get<1>(out), symbol.getSymbol());
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpEpsilon < alphabet::Symbol > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#E";
}

void RegExpToStringComposer::Formal::visit( const FormalRegExpEmpty < alphabet::Symbol > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#0";
}

void RegExpToStringComposer::compose(std::ostream& output, const RegExp& regexp) {
	dispatch(output, regexp.getData());
}

} /* namespace regexp */
