/*
 * TreeFromStringParser.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "TreeFromStringParser.h"
#include "exception/CommonException.h"
#include "tree/TreeClasses.h"
#include <tree/Tree.h>

#include "../StringApi.hpp"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/NonlinearVariableSymbol.h"

namespace tree {

Tree TreeFromStringParser::parseTree ( std::istream & input ) {
	return parseTree ( input, std::set < FEATURES > ( { FEATURES::RANKED_TREE, FEATURES::RANKED_PATTERN, FEATURES::RANKED_NONLINEAR_PATTERN, FEATURES::UNRANKED_TREE, FEATURES::UNRANKED_PATTERN, FEATURES::UNRANKED_NONLINEAR_PATTERN } ) );
}

Tree TreeFromStringParser::parseTree ( std::istream & input, const std::set < FEATURES > & features ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
			return Tree ( UnrankedPattern < > ( subtreeWildcard, std::tree < alphabet::Symbol > ( subtreeWildcard, { } ) ) );
		} else {
			std::ranked_symbol < > subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
			return Tree ( RankedPattern < > ( subtreeWildcard, std::tree < std::ranked_symbol < > > ( subtreeWildcard, { } ) ) );
		}
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::BAR ) {
			alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
			alphabet::Symbol nonlinearVariable ( alphabet::NonlinearVariableSymbol < > ( alphabet::symbolFrom ( token.value ) ) );
			return Tree ( UnrankedNonlinearPattern < > ( subtreeWildcard, { nonlinearVariable }, std::tree < alphabet::Symbol > ( subtreeWildcard, { } ) ) );
		} else {
			std::ranked_symbol < > subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
			std::ranked_symbol < > nonlinearVariable ( alphabet::Symbol ( alphabet::NonlinearVariableSymbol < > ( alphabet::symbolFrom ( token.value ) ) ), 0 );
			return Tree ( RankedNonlinearPattern < > ( subtreeWildcard, { nonlinearVariable }, std::tree < std::ranked_symbol < > > ( nonlinearVariable, { } ) ) );
		}
	} else {
		TreeFromStringLexer::putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;

		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK ) {
			std::set < std::ranked_symbol < > > nonlinearVariables;

			rank = std::from_string < unsigned > ( token.value );

			std::vector < std::tree < std::ranked_symbol < > > > childs;
			bool isPattern = false;

			for ( unsigned i = 0; i < rank; i++ )
				childs.emplace_back ( parseRankedContent ( input, isPattern, nonlinearVariables ) );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				std::ranked_symbol < > subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
				RankedNonlinearPattern < > tree ( std::move ( subtreeWildcard ), std::move ( nonlinearVariables ), std::tree < std::ranked_symbol < > > ( std::ranked_symbol < > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				std::ranked_symbol < > subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
				RankedPattern < > tree ( std::move ( subtreeWildcard ), std::tree < std::ranked_symbol < > > ( std::ranked_symbol < > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				RankedTree < > tree ( std::tree < std::ranked_symbol < > > ( std::ranked_symbol < > ( std::move ( symbol ), rank ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::RANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		} else {
			std::set < alphabet::Symbol > nonlinearVariables;

			std::vector < std::tree < alphabet::Symbol > > childs;
			bool isPattern = false;

			while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
				TreeFromStringLexer::putback ( input, token );
				childs.emplace_back ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) );
				token = TreeFromStringLexer::next ( input );
			}

			if ( token.type != TreeFromStringLexer::TokenType::BAR )
				throw exception::CommonException ( "Missing bar" );

			if ( isPattern && nonlinearVariables.size ( ) ) {
				alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
				UnrankedNonlinearPattern < > tree ( std::move ( subtreeWildcard ), std::move ( nonlinearVariables ), std::tree < alphabet::Symbol > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_NONLINEAR_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else if ( isPattern ) {
				alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
				UnrankedPattern < > tree ( std::move ( subtreeWildcard ), std::tree < alphabet::Symbol > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_PATTERN ) ) return Tree ( std::move ( tree ) );
			} else {
				UnrankedTree < > tree ( std::tree < alphabet::Symbol > ( std::move ( symbol ), std::move ( childs ) ) );

				if ( features.count ( FEATURES::UNRANKED_TREE ) ) return Tree ( std::move ( tree ) );
			}
		}

		throw exception::CommonException ( "Invalid input" );
	}
}

std::tree < std::ranked_symbol < > > TreeFromStringParser::parseRankedContent ( std::istream & input, bool & isPattern, std::set < std::ranked_symbol < > > & nonlinearVariables ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		isPattern = true;
		std::ranked_symbol < > subtreeWildcard ( alphabet::Symbol ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD ), 0 );
		return std::tree < std::ranked_symbol < > > ( std::move ( subtreeWildcard ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		isPattern = true;
		std::ranked_symbol < > nonlinearVariable ( alphabet::Symbol ( alphabet::NonlinearVariableSymbol < > ( alphabet::symbolFrom ( token.value ) ) ), 0 );
		nonlinearVariables.insert ( nonlinearVariable );
		return std::tree < std::ranked_symbol < > > ( std::move ( nonlinearVariable ), { } );
	} else {
		TreeFromStringLexer::putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		unsigned rank = 0;
		std::vector < std::tree < std::ranked_symbol < > > > childs;

		token = TreeFromStringLexer::next ( input );

		if ( token.type == TreeFromStringLexer::TokenType::RANK )
			rank = std::from_string < unsigned > ( token.value );
		else
			throw exception::CommonException ( "Missing rank" );

		for ( unsigned i = 0; i < rank; i++ )
			childs.emplace_back ( std::tree < std::ranked_symbol < > > ( parseRankedContent ( input, isPattern, nonlinearVariables ) ) );

		return std::tree < std::ranked_symbol < > > ( std::ranked_symbol < > ( std::move ( symbol ), rank ), std::move ( childs ) );
	}
}

std::tree < alphabet::Symbol > TreeFromStringParser::parseUnrankedContent ( std::istream & input, bool & isPattern, std::set < alphabet::Symbol > & nonlinearVariables ) {
	TreeFromStringLexer::Token token = TreeFromStringLexer::next ( input );

	if ( token.type == TreeFromStringLexer::TokenType::SUBTREE_WILDCARD ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		alphabet::Symbol subtreeWildcard ( alphabet::SubtreeWildcardSymbol::SUBTREE_WILDCARD );
		return std::tree < alphabet::Symbol > ( std::move ( subtreeWildcard ), { } );
	} else if ( token.type == TreeFromStringLexer::TokenType::NONLINEAR_VARIABLE ) {
		token = TreeFromStringLexer::next ( input );

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		isPattern = true;
		alphabet::Symbol nonlinearVariable ( alphabet::NonlinearVariableSymbol < > ( alphabet::symbolFrom ( token.value ) ) );
		nonlinearVariables.insert ( nonlinearVariable );
		return std::tree < alphabet::Symbol > ( std::move ( nonlinearVariable ), { } );
	} else {
		TreeFromStringLexer::putback ( input, token );
		alphabet::Symbol symbol = alib::stringApi < alphabet::Symbol >::parse ( input );

		std::vector < std::tree < alphabet::Symbol > > childs;

		token = TreeFromStringLexer::next ( input );

		while ( token.type != TreeFromStringLexer::TokenType::BAR ) {
			TreeFromStringLexer::putback ( input, token );
			childs.emplace_back ( parseUnrankedContent ( input, isPattern, nonlinearVariables ) );
			token = TreeFromStringLexer::next ( input );
		}

		if ( token.type != TreeFromStringLexer::TokenType::BAR )
			throw exception::CommonException ( "Missing bar" );

		return std::tree < alphabet::Symbol > ( std::move ( symbol ), std::move ( childs ) );
	}
}

} /* namespace tree */
